#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012,2017 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ctypes
try:
    _libpycsa_handle_ = ctypes.CDLL ('/usr/local/lib/libpycsa.so')
except OSError:
    None

from .version import __version__
from .connset import Mask
from .connset import ConnectionSet
from .elementary import *
#from operators import *
#from arithmetic import *
from .misc import *
from .geometry import *
from .csaobject import parse, from_xml
from .plot import *
from .closure import *
from .conngen import *
